#include "FdIO.h"

/*-------------------------------------------------------private functions---*/
typedef struct FdStat FdStat;

struct FdStat{
  char*      suffix;         /* channel name suffix */
  char*      serName;        /* name of the serData*/
  FdAction*  action;         /* action object which hold this recast object */
};

/*---------------------------------------------------------------------------*/
void FdStatBuild(FdIO *fdIO, FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  FrAdcData *adc;
  FrSerData *sms;
  FdAction *action;
  FrProcData *proc;
  int nAdc, nSms, nProc, len, nImages, adcSize;
  double latency;
  char *p, *serData, *userInfo;

  latency = FrameLatency(frame);

  /*--------------------------------------- compute the number of channels---*/
  nAdc    = 0;
  nImages = 0;
  nSms    = 0;
  nProc   = 0;
  adcSize = 0;

  if(frame->rawData != NULL)  {
    for(adc = frame->rawData->firstAdc; adc != NULL; adc = adc->next) {
       if(adc->data == NULL) continue;
       if(adc->data->nDim == 1) nAdc++;
       else                     nImages++;}
    for(sms = frame->rawData->firstSer; sms != NULL; sms = sms->next) nSms++;
    adcSize = FrameGetAdcSize(frame);}

  for(proc = frame->procData; proc != NULL; proc = proc->next) nProc++;

  /*----------------------------------------------allocate FrSerData space---*/
  len = 1024;

  for(action = fdIO->actionsIn; action != NULL; action = action->next) {
     if(action->serData != NULL) len += strlen(action->serData) + 2;}
  for(action = fdIO->actionsOut; action != NULL; action = action->next) {
     if(action->serData != NULL) len += strlen(action->serData) + 2;}

  serData = malloc(len);
  if(serData == NULL) CfgMsgAddFatal("Malloc serData failed");

  /*---------------------------------------------fill the FrSerData buffer---*/
  p = serData;
  p += sprintf(p,"latency %.2f nAdc %d nImages %d nSms %d nProc %d adcSize %d",
	       latency, nAdc, nImages, nSms, nProc, adcSize);

 for(action = fdIO->actionsIn; action != NULL; action = action->next) {
   if(action->serData != NULL) p += sprintf(p," %s",action->serData);}
 for(action = fdIO->actionsOut; action != NULL; action = action->next) {
   if(action->serData != NULL) p += sprintf(p," %s",action->serData);}

  if(fdIO->actionsIn != NULL) {
    p += sprintf(p," inTime %.4f", fdIO->inputTime);}

  if(fdIO->actionsOut != NULL) {
    p += sprintf(p," outTime %.4f", fdIO->outputTime);}

  if(fdIO->serData != NULL) free(fdIO->serData);
  fdIO->serData = serData;

  /*------------------------------------------------allocate userInfo space---*/
  len = 1024;

  for(action = fdIO->actionsIn; action != NULL; action = action->next) {
     if(action->userInfo != NULL) len += strlen(action->userInfo) + 2;}
  for(action = fdIO->actionsOut; action != NULL; action = action->next) {
     if(action->userInfo != NULL) len += strlen(action->userInfo) + 2;}

  userInfo = malloc(len);
  if(userInfo == NULL) CfgMsgAddFatal("Malloc userInfo failed");

  /*----------------------------------------------fill the userInfo buffer---*/
  p = userInfo;
  p += sprintf(p,"%d-%d latency %.2f nAdc %d",
	       frame->GTimeS,((int)frame->dt), latency, nAdc);
  if(nSms  > 0) p += sprintf(p," nSms %d", nSms);
  if(nProc > 0) p += sprintf(p," nProc %d", nProc);
  p += sprintf(p,";");

 for(action = fdIO->actionsIn; action != NULL; action = action->next) {
   if(action->userInfo == NULL)     continue;
   if(strlen(action->userInfo) == 0) continue; 
   p += sprintf(p," %s",action->userInfo);}
 for(action = fdIO->actionsOut; action != NULL; action = action->next) {
   if(action->userInfo == NULL)     continue;
   if(strlen(action->userInfo) == 0) continue; 
   p += sprintf(p," %s",action->userInfo);}

  if(fdIO->actionsIn != NULL) {
    p += sprintf(p," inTime %.4f", fdIO->inputTime);}

  if(fdIO->actionsOut != NULL) {
    p += sprintf(p," outTime %.4f", fdIO->outputTime);}

  if(fdIO->userInfo != NULL) free(fdIO->userInfo);
  fdIO->userInfo = userInfo;

  return;
}
/*---------------------------------------------------------------------------*/
void FdStatProcessOne(FdStat *s, FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  double sampleRate;
  FdIO* fdIO;

  /*--------------------------------------------build the statistic string---*/
  fdIO = FdIOGetRoot();
  FdStatBuild(fdIO, frame);
  if(fdIO->userInfo == NULL) return;

  /*----------------------------------------------------build the FrSerData---*/
  sampleRate = 1./frame->dt;
  FrSerDataNew (frame, s->serName, frame->GTimeS, fdIO->serData, sampleRate);

  /*-----------------------------------store it to be used by the CmHandler---*/
  CfgMsgAddUserInfo(fdIO->userInfo);

  return;
}
/*---------------------------------------------------------------------------*/
void FdStatProcess(FdStat *stat, FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  FdAction *next;

  if(frame != NULL) FdStatProcessOne(stat, frame);

  next = stat->action->next;
  if(next != NULL) next->action(next->data, frame);

  return;
}
/*--------------------------------------------------------------------------*/
int FdStatNew(void *arg,   CfgDataType_t *data)
/*--------------------------------------------------------------------------*/
{
  FdStat *s, *sCheck;
  int sLen;
  FRBOOL duplicated;
  FdAction *a;
  FdIO* fdIO;

  s = (FdStat*) calloc(1,sizeof(FdStat));
  if(s == NULL) CfgMsgAddFatal("malloc FdStat failed");

  FrStrCpy(&(s->suffix), CfgParseGetNextString(data));

  /*--------------------------------------------------build FrSerData name---*/
  fdIO = FdIOGetRoot();
  sLen = strlen(fdIO->chPrefix) + strlen(CfgGetCmName())+2;
  if(s->suffix != NULL) sLen += strlen(s->suffix);
  s->serName = malloc(sLen);
  if(s->suffix == NULL) sprintf(s->serName, "%s_%s",   fdIO->chPrefix, CfgGetCmName());
  else                  sprintf(s->serName, "%s_%s%s", fdIO->chPrefix, CfgGetCmName(),s->suffix);

  CfgMsgAddInfo("Will compute statistic (latency, nAdc...)"
		" SerData name:%s ",s->serName);

  /*------------------is there another FDIN/OUT_STAT with the same suffix?---*/
  duplicated = FR_NO;

  for(a = fdIO->actionsIn; a != NULL; a = a->next) {
    if(strcmp(a->type, "Stat") != 0) continue;
    sCheck = a->data;
    if(FrStrcmp(s->suffix, sCheck->suffix) == 0) duplicated = FR_YES;}

  for(a = fdIO->actionsOut; a != NULL; a = a->next) {
    if(strcmp(a->type, "Stat") != 0) continue;
    sCheck = a->data;
    if(FrStrcmp(s->suffix, sCheck->suffix) == 0) duplicated = FR_YES;}

  if(duplicated == FR_YES) 
    CfgMsgAddFatal("Two FDIN/OUT_STAT keys with the same suffix");

  /*------------------add the action to the list after the duplicated test---*/
  s->action = FdActionNew((FdAction**) arg, 
			  (void*) s, FdStatProcess, "Stat");

  return(CFG_OK);
}
/*---------------------------------------------------------------------------*/
void FdStatParserAdd(FdIO* fdIO)
/*---------------------------------------------------------------------------*/
{
 
  CfgParseGetFunctionAdd(fdIO->parser, "FDIN_STAT",
			 FdStatNew, (void *) &(fdIO->actionsIn), 1, 
			 CfgString);
  CfgParseGetFunctionAdd(fdIO->parser, "FDOUT_STAT", 
			 FdStatNew, (void *) &(fdIO->actionsOut), 1,
			 CfgString);

 
  return;
}
